import { buildASTSchema, buildSchema, isSchema, } from 'graphql';
import { createYoga } from 'graphql-yoga';
import { useSupergraph } from '@graphql-mesh/fusion-runtime';
// eslint-disable-next-line import/no-extraneous-dependencies
import { DefaultLogger, getHeadersObj } from '@graphql-mesh/utils';
import { buildHTTPExecutor } from '@graphql-tools/executor-http';
import { useExecutor } from '@graphql-tools/executor-yoga';
import { getStitchedSchemaFromSupergraphSdl } from '@graphql-tools/federation';
import { createGraphQLError, getDocumentNodeFromSchema, isDocumentNode, } from '@graphql-tools/utils';
import { isPromise } from '@whatwg-node/server';
import { httpTransport } from './http-transport';
export function createMeshHTTPHandler(config) {
    let transportImportFn;
    if (config.transports != null) {
        if (typeof config.transports === 'function') {
            transportImportFn = config.transports;
        }
        else {
            transportImportFn = (handlerName) => config.transports[handlerName];
        }
    }
    else {
        transportImportFn = defaultTransportImport;
    }
    let fetchAPI = config.fetchAPI;
    // eslint-disable-next-line prefer-const
    let logger;
    // TODO: Move these to other packages later
    function defaultTransportImport(handlerName) {
        if (handlerName === 'http') {
            return httpTransport;
        }
        if (handlerName === 'rest') {
            const omnigraphPackageName = '@omnigraph/json-schema';
            return import(omnigraphPackageName);
        }
        const omnigraphPackageName = `@omnigraph/${handlerName}`;
        return import(omnigraphPackageName);
    }
    const supergraphSpec = config.spec || 'fusion';
    let supergraphYogaPlugin;
    if (config.http) {
        const executor = buildHTTPExecutor({
            fetch: fetchAPI?.fetch,
            ...config.http,
        });
        supergraphYogaPlugin = {
            ...useExecutor(executor),
            invalidateSupergraph() { },
        };
    }
    else if (supergraphSpec === 'fusion') {
        supergraphYogaPlugin = useSupergraph({
            getSupergraph() {
                return handleSupergraphConfig(config.supergraph, yoga.logger);
            },
            getTransportExecutor(transportEntry, getSubgraph, subgraphName) {
                function handleImportResult(importRes) {
                    const getSubgraphExecutor = importRes.getSubgraphExecutor;
                    if (!getSubgraphExecutor) {
                        throw createGraphQLError(`getSubgraphExecutor is not exported from the transport: ${transportEntry.kind}`);
                    }
                    return getSubgraphExecutor(transportEntry, getSubgraph);
                }
                logger.info(`Loading transport "${transportEntry?.kind}" for subgraph "${subgraphName}"`);
                const importRes$ = transportImportFn(transportEntry.kind);
                if (isPromise(importRes$)) {
                    return importRes$.then(handleImportResult);
                }
                return handleImportResult(importRes$);
            },
            polling: config.polling,
        });
    }
    else if (supergraphSpec === 'federation') {
        let supergraph;
        // eslint-disable-next-line no-inner-declarations
        function getAndSetSupergraph() {
            const newSupergraph$ = handleSupergraphConfig(config.supergraph, yoga.logger);
            if (isPromise(newSupergraph$)) {
                return newSupergraph$
                    .then(newSupergraph => {
                    supergraph = getStitchedSchemaFromSupergraphSdl({
                        supergraphSdl: getDocumentNodeFromSchema(newSupergraph),
                    });
                })
                    .catch(e => {
                    yoga.logger.error(`Failed to load the new Supergraph: ${e.stack}`);
                });
            }
            supergraph = getStitchedSchemaFromSupergraphSdl({
                supergraphSdl: getDocumentNodeFromSchema(newSupergraph$),
            });
        }
        let initialSupergraph$;
        supergraphYogaPlugin = {
            onRequestParse() {
                return {
                    onRequestParseDone() {
                        initialSupergraph$ ||= getAndSetSupergraph();
                        return initialSupergraph$;
                    },
                };
            },
            onEnveloped({ setSchema }) {
                setSchema(supergraph);
            },
            invalidateSupergraph() {
                return getAndSetSupergraph();
            },
        };
    }
    const yoga = createYoga({
        fetchAPI: config.fetchAPI,
        logging: config.logging == null ? new DefaultLogger() : config.logging,
        plugins: [...(config.plugins || []), supergraphYogaPlugin],
        context({ request, req, connectionParams }) {
            // Maybe Node-like environment
            if (req?.headers) {
                return {
                    fetch: fetchAPI.fetch,
                    logger,
                    headers: getHeadersObj(req.headers),
                    connectionParams,
                };
            }
            // Fetch environment
            if (request?.headers) {
                return {
                    fetch: fetchAPI.fetch,
                    logger,
                    headers: getHeadersObj(request.headers),
                    connectionParams,
                };
            }
            return {};
        },
        cors: config.cors,
        graphiql: config.graphiql,
        batching: config.batching,
    });
    fetchAPI ||= yoga.fetchAPI;
    logger = yoga.logger;
    Object.defineProperty(yoga, 'invalidateSupergraph', {
        value: supergraphYogaPlugin.invalidateSupergraph,
        configurable: true,
    });
    return yoga;
}
function handleSupergraphConfig(supergraphConfig, logger) {
    if (isPromise(supergraphConfig)) {
        return supergraphConfig.then(newConfig => handleSupergraphConfig(newConfig, logger));
    }
    if (typeof supergraphConfig === 'function') {
        return handleSupergraphConfig(supergraphConfig(), logger);
    }
    if (isSchema(supergraphConfig)) {
        return supergraphConfig;
    }
    if (typeof supergraphConfig === 'string') {
        try {
            return buildSchema(supergraphConfig, {
                assumeValid: true,
                assumeValidSDL: true,
            });
        }
        catch (e) {
            logger.error(`Failed to load Supergraph from ${supergraphConfig}`);
            throw e;
        }
    }
    if (isDocumentNode(supergraphConfig)) {
        return buildASTSchema(supergraphConfig, {
            assumeValid: true,
            assumeValidSDL: true,
        });
    }
    throw new Error(`Invalid Supergraph config: ${supergraphConfig}. It can be an SDL string, a GraphQLSchema, a DocumentNode or a function that returns any of these`);
}
